import crossfilter from 'crossfilter';
import { ascending } from 'd3-array';
import { nest } from 'd3-collection';
import { count, format, groupby, max, mean, median, min, quartile, stdev, sum, variance } from 'datalib';
import { clone, cloneDeep, concat, difference, forEach, has, keys, map } from 'lodash';

var PivotData = /** @class */ (function () {
    function PivotData(pivotData, filters, unitMap) {
        this.startDateTs = 0;
        this.endDateTs = 0;
        this.chartTypeCurrent = '';
        this.selectedRowsCols = [[], [], ['']];
        this.calculateValues = true;
        this.currentAggregator = '';
        this.pivotData = [];
        this.filteredDataSet = [];
        this.rowAttrs = [];
        this.colAttrs = [];
        this.inclusions = new Map();
        this.hiddenAttrs = ['id', 'timestamp', 'Value'];
        this.opts = {
            // showTotals: true,
            readOnly: true,
            header: {
                className: 'header'
            },
            total: {
                className: 'total',
                format: '0,0.00'
            },
            data: {
                className: 'data',
                format: '0,0.00'
            },
            sorted: {
                className: 'sorted',
                format: '0,0.00'
            }
        };
        this.unitMap = new Map();
        if (filters !== undefined) {
            this.inclusions = this.objToStrMap(filters);
        }
        if (unitMap !== undefined) {
            this.unitMap = unitMap;
        }
        this.setData(pivotData);
    }
    // general
    PivotData.prototype.filterDataSet = function (isTime, sortBy) {
        if (isTime === void 0) { isTime = true; }
        if (sortBy === void 0) { sortBy = 'id'; }
        var mainDataSet = crossfilter(this.pivotData);
        var main = mainDataSet.dimension(function (d) { return d[sortBy] || ''; });
        var timeStamp = mainDataSet.dimension(function (d) { return d.timestamp; });
        this.inclusions.forEach(function (value, fields) {
            value = value.map(function (item) { return String(item); });
            if (!fields.includes('Aggregator')) {
                var dimension = mainDataSet.dimension(function (d) { return d[fields] || []; });
                dimension.filterFunction(function (d) { return value.indexOf(String(d)) > -1; });
            }
        });
        return timeStamp.bottom(Infinity);
    };
    // general
    PivotData.prototype.groupDataSet = function (operation, attribute, customFunc, timeLineAttr) {
        if (operation === void 0) { operation = 'count'; }
        if (customFunc === void 0) { customFunc = false; }
        var groupOnMe = concat(this.rowAttrs, this.colAttrs);
        if (timeLineAttr !== undefined && timeLineAttr.length > 0) {
            groupOnMe.push(timeLineAttr);
        }
        var data = this.filterDataSet();
        if (operation === 'count') {
            return groupby(groupOnMe)
                .count()
                .execute(data);
        }
        if (customFunc) {
            return groupby(groupOnMe).execute(data);
        }
        else {
            return groupby(groupOnMe)
                .summarize([{ name: attribute, ops: [operation] }])
                .execute(data);
        }
    };
    // general
    PivotData.prototype.getKeysAttrsV3 = function (attrs, attrToGroup, keys$$1, data) {
        if (attrToGroup === void 0) { attrToGroup = ''; }
        if (keys$$1 === void 0) { keys$$1 = []; }
        if (data === void 0) { data = this.filteredDataSet; }
        var dayArray = ['Sun', 'Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat'];
        var nestedData = nest();
        forEach(attrs, function (attr) {
            if (attr === 'Day') {
                nestedData = nestedData.key(function (d) { return d[attr]; }).sortKeys(function (a, b) {
                    var indexA = dayArray.indexOf(a);
                    var indexB = dayArray.indexOf(b);
                    if (indexA === indexB) {
                        return 0;
                    }
                    else if (indexA < indexB) {
                        return -1;
                    }
                    else {
                        return 1;
                    }
                });
            }
            else {
                nestedData = nestedData.key(function (d) { return d[attr]; }).sortKeys(ascending);
            }
        });
        var crunchedData = nestedData.rollup(function (d) { return d['Value']; }).entries(data);
        return this.retrieveValue(crunchedData);
    };
    // general
    PivotData.prototype.retrieveValue = function (array, element) {
        var _this = this;
        if (element === void 0) { element = []; }
        var returnMe = [];
        forEach(array, function (val) {
            if (val.values !== undefined) {
                element.push(val.key);
                var returnedKeys = _this.retrieveValue(val.values, element);
                returnMe = concat(returnMe, returnedKeys);
            }
            else {
                var newEl = clone(element);
                newEl.push(val.key);
                returnMe.push(newEl);
            }
        });
        element.pop();
        return returnMe;
    };
    // general
    PivotData.prototype.setFilterOnAttr = function (attr, selected) {
        if (selected === undefined) {
            this.inclusions.delete(attr);
        }
        else {
            this.inclusions.set(attr, selected);
        }
        this.filteredDataSet = this.filterDataSet();
        this.calculateValues = true;
    };
    // general
    PivotData.prototype.getRowKeys = function () {
        return this.getKeysAttrsV3(this.rowAttrs);
        // return this.getKeysAttrs(this.rowAttrs);
    };
    // general
    PivotData.prototype.getColKeys = function () {
        return this.getKeysAttrsV3(this.colAttrs);
        // return this.getKeysAttrs(this.colAttrs);
    };
    // general
    PivotData.prototype.getKeys = function () {
        if (this.pivotData.length > 0) {
            var element = this.pivotData[0];
            this.allAttrs = keys(element);
        }
    };
    Object.defineProperty(PivotData.prototype, "allAttr", {
        // general
        get: function () {
            return concat(difference(this.allAttrs, this.hiddenAttrs), [
                'Aggregator'
            ]);
        },
        enumerable: true,
        configurable: true
    });
    // general
    PivotData.prototype.setRowAttrs = function (rowAttrs) {
        this.rowAttrs = rowAttrs;
        this.calculateValues = this.refreshNeeded(this.calculateValues, rowAttrs, true);
    };
    // general
    PivotData.prototype.setColAttrs = function (colAttrs) {
        this.colAttrs = colAttrs;
        this.calculateValues = this.refreshNeeded(this.calculateValues, colAttrs, false);
    };
    // general
    PivotData.prototype.refreshNeeded = function (initState, newArray, isRow) {
        if (isRow === void 0) { isRow = false; }
        if (initState) {
            return initState;
        }
        else {
            var existingData = isRow
                ? this.selectedRowsCols[0]
                : this.selectedRowsCols[1];
            if (existingData.length !== newArray.length) {
                return true;
            }
            else {
                return difference(newArray, existingData).length > 0;
            }
        }
    };
    // general
    PivotData.prototype.getRowAttrs = function () {
        return this.rowAttrs;
    };
    // general
    PivotData.prototype.getColAttrs = function () {
        return this.colAttrs;
    };
    // general
    PivotData.prototype.getData = function () {
        return this.pivotData;
    };
    // general
    PivotData.prototype.getKeysOneAttrV2 = function (attr, getAll) {
        if (getAll === void 0) { getAll = false; }
        if (attr.includes('Aggregator') && !getAll) {
            if (this.inclusions.has('Aggregator')) {
                var returnMe = this.inclusions.get('Aggregator');
                if (returnMe === undefined) {
                    return [];
                }
                return returnMe;
            }
            return this.getAgg();
        }
        else {
            if (this.inclusions.has(attr) && !getAll) {
                var returnMe = this.inclusions.get(attr);
                if (returnMe === undefined) {
                    return [];
                }
                return returnMe;
            }
            return groupby([attr])
                .count()
                .execute(this.pivotData)
                .map(function (item) { return item[attr]; });
        }
    };
    // general
    PivotData.prototype.isFilterApplied = function (attr) {
        return this.inclusions.has(attr);
    };
    // general
    PivotData.prototype.getAgg = function () {
        var opDict = {
            Average: 'mean',
            Count: 'count',
            Sum: 'sum',
            Maximum: 'max',
            Minimum: 'min',
            Median: 'median',
            Variance: 'variance',
            '25 percentile': 'q1',
            '75 percentile': 'q3',
            'Standard deviation': 'stdev',
            'IQR (q3-q1)': 'iqr'
        };
        return keys(opDict);
    };
    // General
    PivotData.prototype.getAppropriateTotalBareBone = function (array, operation, attribute) {
        var format$$1 = format.number(',.2f');
        var opDict = {
            Count: 'count',
            Sum: 'sum',
            Average: 'mean',
            Maximum: 'max',
            Minimum: 'min',
            Median: 'median',
            Variance: 'variance',
            '25 percentile': 'q1',
            '75 percentile': 'q3',
            'Standard deviation': 'stdev',
            'IQR (q3-q1)': 'iqr',
            BoxPlot: 'boxplot'
        };
        var allData = array;
        var result = 0;
        var resultTemp = [];
        switch (opDict[operation]) {
            case 'count':
                result = count(allData, attribute);
                break;
            case 'sum':
                result = sum(allData, attribute);
                break;
            case 'distinct':
                result = count.distinct(allData, attribute);
                break;
            case 'median':
                result = median(allData, attribute);
                break;
            case 'variance':
                result = variance(allData, attribute);
                break;
            case 'stdev':
                result = stdev(allData, attribute);
                break;
            case 'q1':
                resultTemp = quartile(allData, attribute);
                result = resultTemp[0];
                break;
            case 'q3':
                resultTemp = quartile(allData, attribute);
                result = resultTemp[2];
                break;
            case 'iqr':
                resultTemp = quartile(allData, attribute);
                result = resultTemp[2] - resultTemp[0];
                break;
            case 'min':
                result = min(allData, attribute);
                break;
            case 'max':
                result = max(allData, attribute);
                break;
            case 'mean':
                result = mean(allData, attribute);
                break;
            case 'boxplot':
                result = [
                    min(allData, attribute)
                ].concat(quartile(allData, attribute), [
                    max(allData, attribute)
                ]);
                result = result.map(function (val) { return Number(val.toFixed(2)); });
                break;
            default:
                break;
        }
        if (typeof result === 'number') {
            return isNaN(result) ? NaN : Number(result.toFixed(2));
        }
        else {
            return result;
        }
    };
    // Widget
    PivotData.prototype.getWidgetsWithData = function (attrs, attrsWithAgg) {
        var _this = this;
        return new Promise(function (resolve, reject) {
            if (attrsWithAgg !== undefined && attrsWithAgg.length > 0) {
                _this.selectedRowsCols = [
                    attrsWithAgg.map(function (i) { return i.attr; }),
                    attrsWithAgg.map(function (i) { return i.agg; }),
                    [''],
                    ['']
                ];
                var returnMe_1 = [];
                forEach(groupby(['Metric']).execute(_this.filterDataSet()), function (val) {
                    var value = _this.doANestedThing(attrsWithAgg, val['values']);
                    val['valueAgg'] = value.toFixed(2);
                    val['values'] = [];
                    returnMe_1.push(val);
                });
                return resolve(returnMe_1);
            }
            var noOfTime = _this.getKeysOneAttrV2('Time').length;
            var noOfDate = _this.getKeysOneAttrV2('Date').length;
            var returnData = [];
            var groupedData = [];
            if (attrs[0] === 'Aggregator') {
                _this.selectedRowsCols = [_this.rowAttrs, _this.colAttrs, [''], ['']];
                groupedData = _this.groupDataSet('custom', 'custom', true);
                var aggList_1 = _this.getKeysOneAttrV2('Aggregator');
                forEach(groupedData, function (oneGroup) {
                    forEach(aggList_1, function (agg) {
                        var newElem = cloneDeep(oneGroup);
                        newElem['Aggregator'] = agg;
                        newElem['valueAgg'] = _this.getAppropriateTotalBareBone(newElem['values'], agg, 'Value');
                        newElem['values'] = _this.quickGroupMap(newElem['values'], noOfDate > noOfTime ? 'Date' : 'Time');
                        returnData.push(newElem);
                    });
                });
                return resolve(returnData);
            }
            else {
                var agg_1 = _this.getAgg()[0];
                if (_this.inclusions.get('Aggregator') !== undefined) {
                    var inclusionAgg = _this.inclusions.get('Aggregator');
                    if (inclusionAgg !== undefined) {
                        agg_1 = inclusionAgg[0];
                    }
                }
                _this.selectedRowsCols = [_this.rowAttrs, _this.colAttrs, [''], [agg_1]];
                groupedData = _this.groupDataSet('custom', 'custom', true);
                forEach(groupedData, function (oneGroup) {
                    oneGroup['Aggregator'] = agg_1;
                    oneGroup['valueAgg'] = _this.getAppropriateTotalBareBone(oneGroup['values'], agg_1, 'Value');
                    oneGroup['values'] = _this.quickGroupMap(oneGroup['values'], noOfDate > noOfTime ? 'Date' : 'Time');
                });
                return resolve(groupedData);
            }
        });
    };
    // general
    PivotData.prototype.doANestedThing = function (attrs, dataSet) {
        if (dataSet === void 0) { dataSet = this.filterDataSet(); }
        var result = nest();
        forEach(attrs, function (opToDo) {
            result = result.key(function (d) { return d[opToDo.attr]; });
        });
        var newResult = result.entries(dataSet);
        return this.reduceNest(newResult, attrs, -1);
    };
    // general
    PivotData.prototype.reduceNest = function (inputResult, attrs, index) {
        var _this = this;
        if (index === void 0) { index = 0; }
        if (inputResult.length > 0) {
            if (has(inputResult[0], 'key') && !has(inputResult[0], 'value')) {
                index += 1;
                var someShot = map(inputResult, function (singleGroup) {
                    singleGroup.value = _this.reduceNest(singleGroup.values, attrs, index);
                    delete singleGroup.values;
                    return singleGroup;
                });
                return this.reduceNest(someShot, attrs, index);
            }
            else {
                var attToDo = attrs[index];
                var agg = attToDo.agg;
                var crunchedValue = void 0;
                if (has(inputResult[0], 'Value')) {
                    crunchedValue = this.getAppropriateTotalBareBone(inputResult, agg, 'Value');
                }
                else {
                    crunchedValue = this.getAppropriateTotalBareBone(inputResult, agg, 'value');
                }
                return crunchedValue;
            }
        }
    };
    // Widget
    PivotData.prototype.quickGroupMap = function (inputArray, param, operation) {
        if (operation === void 0) { operation = 'mean'; }
        return map(groupby([param])
            .summarize([{ name: 'Value', ops: [operation] }])
            .execute(inputArray), function (val) {
            return [val[param], val[operation + '_Value']];
        });
    };
    // general
    PivotData.prototype.isValid = function () {
        return this.pivotData.length > 0;
    };
    // general
    PivotData.prototype.getChosenFilters = function (renderer) {
        if (renderer === void 0) { renderer = ''; }
        var filters = this.strMapToObj(this.inclusions);
        if (filters === false) {
            return 'Empty Filter present';
        }
        var singleAxisSupport = ['pie', 'boxplot', 'single-value'];
        if (singleAxisSupport.indexOf(renderer) === -1 &&
            this.colAttrs.length === 0) {
            // return 'No Column Chosen';
        }
        var finalOptions = {};
        finalOptions['rows'] = this.selectedRowsCols[0];
        finalOptions['cols'] = this.selectedRowsCols[1];
        finalOptions['filters'] = filters;
        finalOptions['aggregator'] = this.selectedRowsCols[3];
        return finalOptions;
    };
    // General
    PivotData.prototype.setData = function (val) {
        this.pivotData = val;
        this.filteredDataSet = val;
        if (this.unitMap.size === 1) {
            if (this.hiddenAttrs.indexOf('Metric') === -1) {
                this.hiddenAttrs.push('Metric');
            }
        }
        else {
            if (this.hiddenAttrs.indexOf('Metric') !== -1) {
                this.hiddenAttrs.splice(this.hiddenAttrs.indexOf('Metric'), 1);
            }
        }
        this.getKeys();
        this.setStartEndDate();
        this.calculateValues = true;
    };
    // General
    PivotData.prototype.setUnitMap = function (newMap) {
        this.unitMap = newMap;
    };
    // Table
    PivotData.prototype.sortIndicatorRenderer = function (cellProperties) {
        cellProperties.readOnly = true;
        cellProperties.className = this.opts.sorted.className;
        return cellProperties;
    };
    // Table
    PivotData.prototype.headerRenderer = function (cellProperties) {
        cellProperties.readOnly = true;
        cellProperties.className = this.opts.header.className;
        return cellProperties;
    };
    // Table
    PivotData.prototype.totalRenderer = function (cellProperties) {
        cellProperties.format = this.opts.total.format;
        cellProperties.className = this.opts.total.className;
        return cellProperties;
    };
    // Table
    PivotData.prototype.dataRenderer = function (cellProperties) {
        cellProperties.readOnly = this.opts.readOnly;
        cellProperties.className = this.opts.data.className;
        cellProperties.format = this.opts.data.format;
        return cellProperties;
    };
    // General
    PivotData.prototype.strMapToObj = function (strMap) {
        var obj = {};
        var nullFound = false;
        strMap.forEach(function (value, key) {
            if (value.length === 0 &&
                typeof value !== 'boolean' &&
                typeof value !== 'string') {
                nullFound = true;
            }
            else {
                obj[key] = value;
            }
        });
        return nullFound ? false : obj;
    };
    // General
    PivotData.prototype.objToStrMap = function (object) {
        var map$$1 = new Map();
        forEach(object, function (value, key) {
            if (isNaN(Number(key))) {
                map$$1.set(key, value);
            }
            else {
                map$$1.set(Number(key), value);
            }
        });
        return map$$1;
    };
    // Widget
    PivotData.prototype.setStartEndDate = function () {
        var mainDataSet = crossfilter(this.pivotData);
        var timeStamp = mainDataSet.dimension(function (d) { return d.timestamp; });
        this.startDateTs = timeStamp.bottom(1)[0]['timestamp'];
        timeStamp.filterAll();
        this.endDateTs = timeStamp.top(1)[0]['timestamp'];
    };
    return PivotData;
}());

export default PivotData;
//# sourceMappingURL=zen-crunch-core.es5.js.map
